import time

from BestArbitrage.BestArbitrage import core
from BestArbitrage.BestArbitrage.intra_exchange.bestchain import MinMax
from BestArbitrage.BestArbitrage.intra_exchange.parser import ArbitrageFinder


class Robot(object):
    def __init__(self, client: core.ClientExchangeData = None, quote_in_account='USDT', use_quote_only=False):
        # noinspection PyTypeChecker
        self.current_chain: MinMax = None
        self.finder = ArbitrageFinder(client=client)
        self.current_quote = quote_in_account
        self.enable_pairs = list(client.client.fetch_tickers().keys())
        self.use_only_quote = use_quote_only

    def check_profit(self, min_profit=0.4):
        chain = self.current_chain
        self.current_chain = self.finder.check(chain.pair1, chain.pair2, chain.pair3)
        return self.current_chain.profit >= min_profit

    def buy_for_all_balance(self, symbol):
        if core._VERBOSE:
            print(f"buy {symbol}")
        if not core._TEST:
            amount = self.finder.client.client.fetch_free_balance()[symbol.split('/')[1]]
            price = self.finder.client.get_price(symbol, 'ask')
            self.finder.client.client.create_order(
                symbol=symbol,
                type='market',
                side='buy',
                amount=amount / price,
                price=price
            )

    def sell_for_all_balance(self, symbol):
        if core._VERBOSE:
            print(f"sell {symbol}")
        if not core._TEST:
            amount = self.finder.client.client.fetch_free_balance()[symbol.split('/')[0]]
            price = self.finder.client.get_price(symbol, 'bid')
            self.finder.client.client.create_order(
                symbol=symbol,
                type='market',
                side='sell',
                amount=amount,
                price=price
            )

    def run_buy_sell_order(self, pair=('BTC/USDT', 'ask')):
        if core._VERBOSE:
            print(pair)
        if pair[1] == 'ask':
            self.buy_for_all_balance(pair[0])
        else:
            self.sell_for_all_balance(pair[0])

    def execute_chain(self, sleep_in_deals=0):
        orders = (self.current_chain.pair1,
                  self.current_chain.pair2,
                  self.current_chain.pair3)
        for e, order in enumerate(orders, 1):
            self.run_buy_sell_order(order)
            if e != 3 and sleep_in_deals:
                time.sleep(sleep_in_deals)

    @staticmethod
    def _get_best_chain(chains) -> MinMax:
        tune = lambda ch: ch.profit
        return max(chains, key=tune)

    def find_chain(self, min_profit=0.5):
        self.finder.client.update_tickers()
        if self.current_chain is not None:
            is_ok = self.check_profit(min_profit=min_profit)
        else:
            is_ok = False

        if is_ok:
            if core._VERBOSE:
                print(self.current_chain)
            return self.current_chain
        else:  # finding best chain
            chains = list(self.finder.start_all_checks())
            best = self._get_best_chain(chains=chains)
            if best.profit >= min_profit:
                self.current_chain = best
            else:
                self.current_chain = None
            if core._VERBOSE:
                print(best)
            return self.current_chain

    def start_arbitrage(self,
                        min_profit=0.4,
                        sleep_in_chains=0,
                        sleep_in_deals=0):
        asset1 = list(set(
            map(lambda x: x.split('/')[0], self.enable_pairs)
        ))
        self.finder.pair_arbitrage_generator(
            alt=asset1,
            usable_pairs=self.enable_pairs
        )
        while True:
            self.find_chain(
                min_profit=min_profit
            )
            if self.current_chain is not None:
                if (not self.use_only_quote) or self.current_quote == self.current_chain.get_needed_coin():
                    self.prepair_to_chain()
                    self.execute_chain(
                        sleep_in_deals=sleep_in_deals
                    )

            time.sleep(sleep_in_chains)

    def prepair_to_chain(self):
        coin = self.current_chain.get_needed_coin()
        if coin != self.current_quote:
            symbol = f"{coin}/{self.current_quote}"
            if symbol not in self.enable_pairs:
                self.sell_for_all_balance(f"{self.current_quote}/{coin}")
            else:
                self.buy_for_all_balance(symbol)
            self.current_quote = coin



# -*- coding: utf-8 -*-
import re

from colorama import Fore as fore

from BestArbitrage.BestArbitrage.core import get_percentage


class MinMax(object):
    lower_price_exchange = None
    higher_price_exchange = None
    lower_price = 0
    higher_price = 0
    percentage_profit = 0
    pair: str

    def __init__(self,
                 lower_price_exchange='Binance',
                 higher_price_exchange='Binance',
                 lower_price=1,
                 higher_price=1,
                 pair=None,
                 lower_price_volume=0,
                 higher_price_volume=0):
        self.lower_price_exchange = re.sub(r"[^\w]", '', lower_price_exchange)
        self.higher_price_exchange = higher_price_exchange
        self.lower_price = lower_price
        self.higher_price = higher_price
        self.lower_price_volume = lower_price_volume
        self.higher_price_volume = higher_price_volume
        self.percentage_profit = get_percentage(higher_price, lower_price)
        self.pair = pair

        if self.percentage_profit < 0.3:
            self.color = fore.LIGHTRED_EX
        elif self.percentage_profit < 1:
            self.color = fore.LIGHTYELLOW_EX
        elif self.percentage_profit < 2:
            self.color = fore.GREEN
        elif self.percentage_profit < 3.4:
            self.color = fore.CYAN
        elif self.percentage_profit < 4.5:
            self.color = fore.BLUE
        else:
            self.color = fore.LIGHTBLUE_EX
        self.data = f"buy {self.pair} at {self.lower_price_exchange} for {self.lower_price}, transfer and sell at \
{self.higher_price_exchange} for {self.higher_price} with {round(self.percentage_profit, 2)}% profit (volumes: \
low price: {self.lower_price_volume}, high price: {self.higher_price_volume})"

    def __repr__(self):
        return f"{self.color}{self.data}{fore.RESET}"



# -*- coding: utf-8 -*-

from selenium.common.exceptions import InvalidSessionIdException

import BestArbitrage.BestArbitrage.inter_exchange.bestexchange as bestexchange
from BestArbitrage.BestArbitrage import core


class ArbitrageFounder(core.AskTradingView):

    def get_data(
            self,
            pair,
            your_exchanges=[
                "BINANCE",
                "FTX"
            ],
            exchange_blacklist=["BITTREX", "CAPITALCOM", "CURRENCYCOM"],
            all_exchanges=True,
            min_volumes={
                'BTC': 0.25,
                'ETH': 5,
                'USDT': 10000
            },
            print_oops=True):
        try:
            exc_pair = {}
            self.driver.get(f"https://ru.tradingview.com/symbols/{pair}/markets/")
            exchanges = self.driver.find_elements_by_class_name(
                "tv-data-table__row.tv-data-table__stroke.tv-screener-table__result-row"
            )
            for item in exchanges:
                item_call = item.find_elements_by_class_name(
                    "tv-data-table__cell.tv-screener-table__cell"
                )
                exchange = item_call[0].text
                volume = item_call[6].text

                if volume[-1] == "K":
                    vol = float(volume[:-1]) * 1_000
                elif volume[-1] == "M":
                    vol = float(volume[:-1]) * 1_000_000
                elif volume[-1] == "B":
                    vol = float(volume[:-1]) * 1_000_000_000
                else:
                    vol = float(volume)

                price = float(item_call[1].text)
                for key, val in zip(min_volumes.keys(),
                                    min_volumes.values()):
                    if pair.endswith(key):
                        min_vol_ = val
                if (
                        exchange in your_exchanges or all_exchanges) and vol >= min_vol_ and exchange not in exchange_blacklist:
                    exc_pair[exchange] = {"volume": vol,
                                          "price": price}
            if len(list(exc_pair.keys())) == 0:
                return None
            list_exchanges = []
            for exchange in exc_pair.keys():
                price = exc_pair[exchange]["price"]
                list_exchanges.append({exchange: price})
            get_sort = lambda x: list(x.values())[0]
            get_sort_key = lambda x: list(x.keys())[0]
            minimal = min(list_exchanges, key=get_sort)
            maximal = max(list_exchanges, key=get_sort)
            return bestexchange.MinMax(
                get_sort_key(minimal),
                get_sort_key(maximal),
                get_sort(minimal),
                get_sort(maximal),
                pair=pair,
                higprex_vol=exc_pair[get_sort_key(maximal)]['volume'],
                loprex_vol=exc_pair[get_sort_key(minimal)]['volume']
            )
        except Exception as e:
            if isinstance(e, (KeyboardInterrupt, InvalidSessionIdException)):
                raise e
            if print_oops:
                print('Oops, something went wrong...')
            return None



import re

import ccxt
import selenium.webdriver.support.expected_conditions as ec
from numpy import inf
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait

_TEST = False
_VERBOSE = False


def get_percentage(higher_price, lower_price):
    increase = higher_price - lower_price
    if lower_price == 0:
        percentage_profit = inf
    else:
        percentage_profit = (increase / lower_price) * 100
    return percentage_profit


class ClientExchangeData(object):
    def __init__(self, client: ccxt.Exchange = None):
        self.client = client
        if self.client is not None:
            self.update_tickers()

    def update_tickers(self):
        self._fetch_tickers = self.client.fetch_tickers()

    def get_price(self, symbol, side='ask/bid'):
        ticker = self._fetch_tickers[symbol]
        if side == "ask":
            return float(ticker['ask'])
        else:
            return float(ticker['bid'])


class AskTradingView(object):
    def __init__(self, driver=None):
        self.driver: webdriver.Safari = driver

    def get_all_tv_pairs(self, screener='crypto-screener'):
        self.driver.get(f"https://ru.tradingview.com/{screener}/")

        waiter = WebDriverWait(self.driver, 5)
        items = waiter.until(ec.visibility_of_all_elements_located(
            (By.CLASS_NAME, "tv-screener__symbol.apply-common-tooltip")))
        return set(map(lambda x: x.text, items))

    def get_all_coins(self, screener='cryptocurrencies', pages=3):
        self.driver.get(f"https://ru.tradingview.com/markets/{screener}/prices-all/")

        waiter = WebDriverWait(self.driver, 5)
        items = []
        loadmore = self.driver.find_element_by_class_name("tv-load-more__btn")
        for page in range(pages):
            items.extend(
                waiter.until(
                    ec.visibility_of_all_elements_located(
                        (By.CLASS_NAME,
                         "tv-data-table__row.tv-data-table__stroke."
                         "tv-screener-table__result-row")
                    )
                )
            )
            loadmore.click()

        cryptos = list(
            map(
                lambda x: re.findall(
                    r":.+",
                    x.get_attribute('data-symbol')
                )[0][1:-3],
                items
            )
        )
        return cryptos



from BestArbitrage.BestArbitrage import core
from BestArbitrage.BestArbitrage.intra_exchange.bestchain import MinMax


class ArbitrageFinder:
    def __init__(self, client: core.ClientExchangeData):
        self.client = client

    @staticmethod
    def get_profit(alt_base_ask: float,
                   alt_shit_bid: float,
                   shit_base_bid: float) -> float:
        """

        buy  ALT/BASE
        sell ALT/SHIT
        sell SHIT/BASE

        BASE -> ALT -> SHIT
        """
        if alt_shit_bid == 0:
            return 0
        return core.get_percentage(shit_base_bid, alt_base_ask / alt_shit_bid)

    def pair_arbitrage_generator(self,
                                 # no, i can't use itertools.product. I need to check
                                 base=['BTC', 'USDT'],
                                 alt=['XMR', 'XLM'],
                                 shit=['BTC', 'USDT'],
                                 usable_pairs=[]):
        alt_base = []
        alt_shit = []
        shit_base = []
        for coin in alt:
            for target in base:
                for market in shit:
                    if coin != target and target != market and market != coin:
                        can = True

                        alba = f"{coin}/{target}"
                        alsh = f"{coin}/{market}"
                        shba = f"{market}/{target}"

                        baal = f"{target}/{coin}"
                        shal = f"{market}/{coin}"
                        bash = f"{target}/{market}"

                        if alba in usable_pairs:
                            alba = (alba, 'ask')
                        elif baal in usable_pairs:
                            alba = (baal, 'bid')
                        else:
                            can = False

                        if alsh in usable_pairs:
                            alsh = (alsh, 'bid')
                        elif shal in usable_pairs:
                            alsh = (shal, 'ask')
                        else:
                            can = False

                        if shba in usable_pairs:
                            shba = (shba, 'bid')
                        elif bash in usable_pairs:
                            shba = (bash, 'ask')
                        else:
                            can = False

                        if can:
                            alt_base.append(alba)
                            alt_shit.append(alsh)
                            shit_base.append(shba)
        self.pairs = tuple(zip(alt_base, alt_shit, shit_base))
        return self.pairs

    def check(self, coin, target, market) -> MinMax:

        c = self.client.get_price(coin[0], coin[1])
        t = self.client.get_price(target[0], target[1])
        m = self.client.get_price(market[0], market[1])
        if c == 0 or t == 0 or m == 0:
            return MinMax(coin, target, market)
        c1 = c if coin[1] == 'ask' else 1 / c
        t1 = t if target[1] == 'bid' else 1 / t
        m1 = m if market[1] == 'bid' else 1 / m
        result = self.get_profit(c1, t1, m1)

        return MinMax(coin, target, market, result)

    def start_all_checks(self):
        for pairs in self.pairs:
            yield self.check(*pairs)



# -*- coding: utf-8 -*-
from distutils.core import setup

import setuptools

with open('./README.md') as file:
    long_desc = file.read()
_version_ = '0.4.1'
setup(
    name='BestArbitrage',
    author='Vlad Kochetov',
    author_email='vladyslavdrrragonkoch@gmail.com',
    packages=setuptools.find_packages(),
    version=_version_,
    description='Arbitrage cryptocurrency smartly',
    long_description=long_desc,
    long_description_content_type="text/markdown",
    url='https://github.com/VladKochetov007/BestArbitrage',
    install_requires=[
        'selenium',
        'numpy',
        'colorama',
        'ccxt'
    ],
    download_url=f'https://github.com/VladKochetov007/BestArbitrage/archive/{_version_}.tar.gz',
)



# -*- coding: utf-8 -*-
from typing import Tuple

from colorama import Fore


class MinMax(object):
    def __init__(self,
                 pair1: Tuple[str, str] = ('None', 'ask'),
                 pair2: Tuple[str, str] = ('None', 'bid'),
                 pair3: Tuple[str, str] = ('None', 'bid'),
                 profit=0):
        self.pair1 = pair1
        self.pair2 = pair2
        self.pair3 = pair3
        self.profit = profit

        if pair1[1] == 'ask':
            BS1 = f"buy {pair1[0]}"
        else:
            BS1 = f"sell {pair1[0]}"

        if pair2[1] == 'ask':
            BS2 = f"buy {pair2[0]}"
        else:
            BS2 = f"sell {pair2[0]}"

        if pair3[1] == 'ask':
            BS3 = f"buy {pair3[0]}"
        else:
            BS3 = f"sell {pair3[0]}"
        self.data = f"{BS1}, {BS2}, {BS3} with {round(self.profit, 3)}% profit"

        if profit <= 0:
            self.color = Fore.BLACK
        elif profit < 0.03:
            self.color = Fore.LIGHTRED_EX
        elif profit < 0.1:
            self.color = Fore.LIGHTYELLOW_EX
        elif profit < 0.2:
            self.color = Fore.GREEN
        elif profit < 0.34:
            self.color = Fore.CYAN
        elif profit < 0.45:
            self.color = Fore.BLUE
        else:
            self.color = Fore.LIGHTBLUE_EX

    def __repr__(self):
        return f"{self.color}{self.data}{Fore.RESET}"

    def get_needed_coin(self):
        get = lambda x: self.pair1[0].split('/')[x]
        if self.pair1[1] == 'ask':
            return get(1)
        else:
            return get(0)
